/*
  Note:

  DREQ is changed in normal firmware both in SDI interrupts when the
  stream buffer becomes too full, and whenever the stream buffer becomes
  empty enough.

  Because we have our own SDI handler, the interrupt does not touch DREQ.
  So, we can use DREQ ourselves as we wish and access to it does not need to
  be protected by interrupt Disable()/Enable(). However, the normal decoding
  will still keep raising DREQ when it sees that there is space for new data.

 */
#define USE_HC

#if 0 && defined(VS1011E)
//#define SKIP_ID3V2    /* skip ID3v2 tags */
#define SHUFFLE_PLAY /*12/19 words*/
#define SAVE_POSITION
#define SAVE_VOLUME
//#define NO_FAST_PAUSE /*already default for VS1011*/
#define LOUDNESS_ALWAYS_ON_OFF
#endif


//#define SAVE_VOLUME /*66 words */
//#define SAVE_POSITION /*32 words*/
#if defined(SAVE_VOLUME) || defined(SAVE_POSITION)
#define SAVE_STUFF
#endif

/*
  - subdirectory support

  NEW features for SCI_UI:
  - play mode:
    + normal continuous play
    + pause before song (doubles as pause after reset)
    + loop song
    + report file found / not found and wait for play command
  - find file by number but have the file name readable from memory

  TODO features:

  Extra features
  - L/R VU-meter through SCI
 */

#include <string.h>
#include <stdlib.h>
#include "standalone.h"

auto void play_frame(void);
//void Disable();
//void Enable();
//void Reset();

//#undef putchar
//#undef puts

//#include <stdio.h>


#if defined(VS1011E)
/* SDI and SCI enables are reversed in VS1011 */
#define ENABLE_SCI_DAC_RX ((1<<INT_EN_SDI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#else
#define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#endif
#define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))

//#define USE_DEBUG

#ifdef USE_DEBUG
#undef ENABLE_UI
#undef SCI_UI
#endif

#if 0
#ifdef VS1011E
#define WAIT_FOR_BUTTON /*After poweron or watchdog reset*/
#define NO_VOLUME_CONTROLS
#define LOUDNESS_ALWAYS_ON_OFF
#endif
#endif


//#define TIMEOUT /* timeout when decoding seems to be stuck */
#define FIX_WAV_DECODETIME  /* SCI_UI: advance decodetime for short WAV's */
//#define HDATCHECK 800 /* 21words: give up after 800/2=400kB ~ 3-6sec */
#ifndef NOCHECKFILETYPE
#define CHECKFILETYPE /* vs1011 29 words? check file type according to short-filename suffix */
#endif/*NOCHECKFILETYPE*/
#define KEEP_DRIVERS_OFF /* .. until MMC located */
#define WITH_SUBDIRECTORIES /* support subdirectories */
//#define SHUFFLE_PLAY /*shuffle instead of random +12/19 words*/

void putch(register __a0 int ch);



#ifdef CHECKFILETYPE
#undef CHECKFILETYPE /*temporary undef*/
#define CHECKFILETYPE CheckFileType
#endif


#define CTRL3_UPDATE_VOL        (1<<15)
#define CTRL3_AT_END            (1<<6)
#define CTRL3_NO_NUMFILES       (1<<5)
#define CTRL3_PAUSE_ON          (1<<4)
#define CTRL3_FILE_READY        (1<<3)

#define CTRL3_PLAY_MODE_MASK    (3<<1)
#define CTRL3_PAUSE_AFTER_PLAY  (3<<1)
#define CTRL3_PAUSE_BEFORE_PLAY (2<<1)
#define CTRL3_LOOP_SONG         (1<<1)
#define CTRL3_PLAY_NEXT         (0<<1)

#define CTRL3_RANDOM_PLAY       (1<<0) /* SCI_AICTRL3 */

#ifndef GPIO_IDATA
#define GPIO_IDATA GPIO_DATA
#endif

#ifdef SWITCH_CS2_AND_SCLK
#define SPI_xCS   1 /* GPIO */
#define SPI_CLK   8
#define SPI_MISO  4
#define SPI_MISO_SHIFT -3
#define SPI_xCS2  2
#define SPI_MOSI  1 /* DREQ */
#else
#define SPI_xCS   1 /* GPIO */
#define SPI_CLK   2
#define SPI_MISO  4
#define SPI_xCS2  8
#define SPI_MOSI  1 /* DREQ */
#endif

//#define SPI_WREN  0x06
//#define SPI_WRDI  0x04
//#define SPI_RDSR  0x05
//#define SPI_WRSR  0x01
#define SPI_READ  0x03
//#define SPI_WRITE 0x02

#ifdef USE_DEBUG
static char hex[] = "0123456789ABCDEF";
void puthex(u_int16 d) {
    register i;
    for (i=0;i<4;i++) {
	putchar(hex[(d>>12)&15]);
	d <<= 4;
    }
}
void puthex2(u_int16 d) {
    register i;
    putchar(hex[(d>>4)&15]);
    putchar(hex[(d>>0)&15]);
}
#endif


typedef enum {
    spi_pause=0,
    spi_file,
    spi_seek,
    spi_waitdata,
    spi_data,
    spi_zeros,
    spi_record,
    spi_waitima
} SPISTATE;

SPISTATE spiState, oldSpiState;
u_int16 spiCnt;
#ifdef TIMEOUT
u_int16 timeOut;
#endif

struct SPIX { /* do not touch! ASM uses the structure */
    s_int32 size;
    u_int32 sector;
    u_int16 fragSize; /* TODO: make this 32-bit and multiply fragment list clusters by sectorsPerCluster. */
    __y struct FRAGMENT *fragment;
} spix;


#define 	MMC_GO_IDLE_STATE   0
#define 	MMC_SEND_OP_COND   1
#define 	MMC_SEND_IF_COND   8
#define 	MMC_SEND_CSD   9
#define 	MMC_SEND_CID   10
#define 	MMC_SEND_STATUS   13
#define 	MMC_SET_BLOCKLEN   16
#define 	MMC_READ_SINGLE_BLOCK   17
#define 	MMC_WRITE_BLOCK   24
#define 	MMC_PROGRAM_CSD   27
#define 	MMC_SET_WRITE_PROT   28
#define 	MMC_CLR_WRITE_PROT   29
#define 	MMC_SEND_WRITE_PROT   30
#define 	MMC_TAG_SECTOR_START   32
#define 	MMC_TAG_SECTOR_END   33
#define 	MMC_UNTAG_SECTOR   34
#define 	MMC_TAG_ERASE_GROUP_START   35
#define 	MMC_TAG_ERARE_GROUP_END   36
#define 	MMC_UNTAG_ERASE_GROUP   37
#define 	MMC_ERASE   38
#define         MMC_READ_OCR     58
#define 	MMC_CRC_ON_OFF   59
#define 	MMC_R1_BUSY   0x80
#define 	MMC_R1_PARAMETER   0x40
#define 	MMC_R1_ADDRESS   0x20
#define 	MMC_R1_ERASE_SEQ   0x10
#define 	MMC_R1_COM_CRC   0x08
#define 	MMC_R1_ILLEGAL_COM   0x04
#define 	MMC_R1_ERASE_RESET   0x02
#define 	MMC_R1_IDLE_STATE   0x01
#define 	MMC_STARTBLOCK_READ   0xFE
#define 	MMC_STARTBLOCK_WRITE   0xFE
#define 	MMC_STARTBLOCK_MWRITE   0xFC
#define 	MMC_STOPTRAN_WRITE   0xFD
#define 	MMC_DE_MASK   0x1F
#define 	MMC_DE_ERROR   0x01
#define 	MMC_DE_CC_ERROR   0x02
#define 	MMC_DE_ECC_FAIL   0x04
#define 	MMC_DE_OUT_OF_RANGE   0x04
#define 	MMC_DE_CARD_LOCKED   0x04
#define 	MMC_DR_MASK   0x1F
#define 	MMC_DR_ACCEPT   0x05
#define 	MMC_DR_REJECT_CRC   0x0B
#define 	MMC_DR_REJECT_WRITE_ERROR   0x0D

#if 1
auto u_int16 MmcCommand(register __b0 s_int16 cmd, register __d u_int32 arg);
#else
auto u_int16 MmcCommand(register __b0 s_int16 cmd, register __d u_int32 arg) {
    register __b1 u_int16 t;
#if 1
    puts("Cmd(");
    puthex(cmd);
    puthex(arg>>16);
    puthex(arg);
    puts(")");
#endif
    /* some MMC brands require the extra clock pulses     */
    SpiSendClocks();
    SpiSendClocks();
    USEX(GPIO_ODATA) &= ~SPI_xCS2;

    while (SpiSendReceiveMmc(0xff00,8) != 0xff) {
	/* Unexpected busy signal from MMC */
#ifdef USE_DEBUG
	putchar('u');
	putchar('\n');
#endif
    }
    SpiSendReceiveMmc(cmd, 16); /* top half normally 0xff */
    SpiSendReceiveMmc(arg>>16, 16);
    SpiSendReceiveMmc(arg, 16);
    t = SpiSendReceiveMmc(0x9500, 8);    /* Valid CRC for init, then don't care */

    cmd = 100;               /* timeout value, try max. 100 times */
    while (cmd && (t & 0x80)) { /* R1 response not detected */
	t = SpiSendReceiveMmc(0xff00, 8);
	cmd--;               /* decrease the timeout value */
    }
#ifdef USE_DEBUG
    puthex(t);
    putchar('\n');
#endif
    return t;
}
#endif

#if 1
auto void ReadDiskSector(register __a u_int32 sector);
#else
auto void ReadDiskSector(register __a u_int32 sector) {
    register s_int16 i;
    register u_int16 *buffer = stream_buffer;

#if 0 && defined(USE_DEBUG)
    //puthex(sector>>16);
    puthex(sector);
    puts(" SECTOR");
#endif

#ifdef USE_HC /* In ASM this uses self-modified code */
    MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector&0x7fffffffUL)<<hcShift);
#else
    MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector<<9));
#endif
    do {
	i = SpiSendReceiveMmc(0xff00,8);
#if 0 && defined(USE_DEBUG)
	puthex(i);
	puts(" wait");
#endif
    } while (i == 0xff);

#if 0
    if (i != 0xfe) {
	SpiSendClocks();/*USEX(GPIO_ODATA) |= SPI_xCS2;*/
#ifdef USE_DEBUG
	puts("wfd fail");
#endif
	return;
    }
#endif

    for (i=0; i<512; i+=2) {
	*buffer++ = SpiSendReceiveMmc(0xffff,16);
    }
    SpiSendReceiveMmc(0xffff,16); /* discard crc */

    /* generate some extra SPI clock edges to finish up the command */
    SpiSendClocks();
    SpiSendClocks();
    return;
}
#endif

#if 0
struct Fat {
    byte BS_jmpBoot[3];		/**< 0 x86 Boot Jump Code */
    byte BS_OEMName[8];		/**< 3 Formatter's name, usually "MSWIN4.1" */
    
    word BPB_BytsPerSec;	/**< 11 Bytes per sector (512) */
    byte BPB_SecPerClus;	/**< 13 Sectors per Cluster (1,2,4,8,..,128) */
    word BPB_RsvdSecCnt;	/**< 14 Reserved sectors (1 (32 for FAT32)) */
    byte BPB_NumFATs;		/**< 16 Number of FATs (2) */
    word BPB_RootEntCnt;	/**< 17 FAT12/16 n of root dir entries */
    word BPB_TotSec16;		/**< 19 Old sector count (0 for FAT32) */
    byte BPB_Media;		/**< 21 Media Type (eg 0xF8) */
    word BPB_FATSz16;		/**< 22 Size of one FAT16 in sectors */
    word BPB_SecPerTrk;		/**< 24 Old CHS Sectors Per Track */
    word BPB_NumHeads;		/**< 26 Old CSH Number of Heads */
    u_32 BPB_HiddSec;		/**< 28 n of sectors before this volume */
    u_32 BPB_TotSec32;		/**< 32 New sector count (0 for FAT12/16) */
   
    /** FAT type specific extensions */
    union Extensions{	

      /** FAT12/16 specific extensions to Bios Parameter Block*/
      struct Fat16Specific {	
	byte BS_DrvNum;		/**< 36 DOS INT13 Drive Number (0x80=HD)*/
	byte BS_Reserved1;	/**< 37 For WINNT; Format to 0 */
	byte BS_BootSig;	/**< 38 0x29 if next 3 fields are present */
	byte BS_VolID[4];	/**< 39 Volume ID (usually format datetime) */
	byte BS_VolLab[11];	/**< 43 Volume Label */
	byte BS_FilSysType[8];	/**< 54 Decorative name of fs, eg "FAT16   "*/
      } _16;

      /** FAT32 specific extensions to Bios Parameter Block*/
      struct Fat32Specific {	
        u_32 BPB_FATSz32;	/**< 36 Size of one FAT32 in sectors */
	word BPB_ExtFlags;	/**< 40 Flags; active FAT number etc */
	word BPB_FSVer;		/**< 42 File System Version (0x0000) */
	u_32 BPB_RootClus;	/**< 44 Start cluster of Root Dir. (2) */
	word BPB_FSInfo;	/**< 48 Start sector of FSINFO in Resvd area */
	word BPB_BkBootSec;	/**< 50 Sector in Resvd area for BkBoot (6) */
	byte BPB_Reserved[12];	/**< 52 Reserved, Always 0. */
	byte BS_DrvNum;      	/**< 64 DOS INT13 Drive Number (0x80=HD) */
	byte BS_Reserved1;      /**< 65 For WINNT, Format to 0 */
	byte BS_BootSig;	/**< 66 0x29 if next 3 fields are present */
	byte BS_VolID[4];	/**< 67 Volume ID (usually format datetime) */
	byte BS_VolLab[11];	/**< 71 Volume Label */
	byte BS_FilSysType[8];  /**< 82 Decorative name of fs, eg "FAT32   "*/
      } _32;
    } ext;
  } fat;
#endif

#if 1
auto u_int16 GetByte(register __c0 u_int16 n);
#else
auto u_int16 GetByte(register __c0 u_int16 n) {
    register u_int16 *p = stream_buffer[n>>1];
    if (n & 1) {
	n = *p;
    } else {
	n = *p >> 8;
    }
    return n & 0xff;
}
#endif
#if 1
auto u_int16 GetWord(register __c0 u_int16 n);
#else
auto u_int16 GetWord(register __c0 u_int16 n) {
    return GetByte(n) | (GetByte(n+1)<<8);
}
#endif
#if 1
auto u_int32 GetLong(register __c0 u_int16 n);
#else
auto u_int32 GetLong(register __c0 u_int16 n) {
    return GetWord(n) | ((u_int32)GetWord(n+2)<<16);
}
#endif

extern __y u_int16 fileName[6];
#ifdef FATINFO_IN_Y
__y struct FATINFO fatInfo;
#else
struct FATINFO fatInfo;
#endif

#ifdef USE_DEBUG
void PrintDiskSector(void) {
    register int i;
    register int j;
    for (i=0;i<512;i+=16) {
	puthex(i);
	putchar(' ');
	putchar(' ');
	for (j=0;j<16;j++) {
	    puthex2(GetByte(i+j));
	    putchar(' ');
	}
	putchar(' ');
	for (j=0;j<16;j++) {
	    register int t = GetByte(i+j);
	    if ((t & 0x7f) < 0x20)
		putchar('?');
	    else
		putchar(t);
	}
	putchar('\n');
    }
}
#endif

#if 1 //&& !defined(USE_DEBUG)
auto u_int16 InitFileSystem(void);
#else
auto u_int16 InitFileSystem(void) {
    /* Load MBR -- the first sector on disk */
    ReadDiskSector(0);

#if 0 && defined(USE_DEBUG)
    /* Ok, it should be a MBR sector. Let's verify */
    if (stream_buffer[255] != 0x55aa) {
#ifdef USE_DEBUG
//      puthex(stream_buffer[255]);
	puts("!MBR");
#endif
	return -1; /* sector 0 is not MBR. */
    }
#endif

#if 0 && defined(USE_DEBUG)
    /* This checks that partition 1 is active. Alter code to allow
       other partition configurations. */
    if (GetByte(0x1be) == 0x80) {
    } else {
	puthex(GetByte(0x1be));
	puts(" PF!");
    }
#endif

    /* Now leave MBR and load sector 0 of partition */
    /* check for boot sector instead of partition table */
    if (GetLong(0x36)!= (((long)'1'<<24)|((long)'T'<<16)|((long)'A'<<8)|'F') &&
	GetLong(0x52)!= (((long)'3'<<24)|((long)'T'<<16)|((long)'A'<<8)|'F')
	) {
#ifdef USE_DEBUG
	PrintDiskSector();
	puts("");
#endif
	fatInfo.currentSector = GetLong(0x1c6);
	ReadDiskSector(fatInfo.currentSector);
    }
#ifdef USE_DEBUG
    PrintDiskSector();
#endif

    {
	u_int16 BPB_BlksPerSec;

	/* Determine FAT Type (16/32) */
	/* This should be done better, but it'll do for now. */
	fatInfo.BPB_RootEntCnt = GetWord(17);
	/* FAT32 does not have separate root entries. */
	/* TODO: should detect FAT32 by BPB_FATSz16==0 */
	fatInfo.IS_FAT_32 = fatInfo.BPB_RootEntCnt ? 0 : 1;
	fatInfo.FilSysType = GetWord(54+3); /* 0x3231 for FAT12 */
#ifdef USE_DEBUG
	puthex(fatInfo.FilSysType);
	puts("fsys");
#endif
#if 0
	if (!IS_FAT_32){
	    if (GetByte(54+4)=='2') {/*diskSect.fat.ext._16.BS_FilSysType[4]=='2'*/) {
		fatInfo.IS_FAT_12 = GetByte(54);
	    }
	}
#endif

#if 0 && defined(USE_DEBUG)
	if (fatInfo.IS_FAT_32) {
	    puts("FAT32");
	} else {
	    puthex(fatInfo.BPB_RootEntCnt);
	    puts(" FAT");
	    puthex(GetWord(19));
	    puts(" TSc16");
	}
	puthex(GetWord(54+3)); /* 3231 == '12' 3631= '16' 0000=FAT32 */
	puts(" fst");
#endif
#if 1 && defined(USE_DEBUG)
	putchar('\"');
	{
	    register int i;
	    for (i=0; i<11; i++) {
		if (fatInfo.IS_FAT_32) {
		    putchar(GetByte(71+i));
		} else {
		    putchar(GetByte(43+i));
		}
	    }
	}
	putchar('\"');
	putchar('\n');
#endif
	/* OK, let's calculate */
	/* First, let's get rid of the idea that we have byte addresses
	   in the file system. Nope, let's only deal in physical disk
	   sectors of 512 bytes. First we convert the FAT byter per sector
	   value to "512B disk sectors per fat sector" value. */

	BPB_BlksPerSec = GetWord(11) >> 9;

	/* Then we adjust the Sector per Cluster to mean physical disk
	   sectors. in 99% of the cases it is already so because bytes
	   per sector almost always is 512 in FAT. Maximum cluster size
	   is 65536 bytes (128 disk sectors). */

	fatInfo.fatSectorsPerCluster = GetByte(13)/*BPB_SecPerClus*/ * BPB_BlksPerSec;
#if 0 && defined(USE_DEBUG)
	puthex(fatInfo.fatSectorsPerCluster);
	puts(" FatSPC");
#endif
	fatInfo.fatStart = fatInfo.currentSector +
	    (u_int32)GetWord(14)/*BPB_RsvdSecCnt*/ * BPB_BlksPerSec;
#if 0 && defined(USE_DEBUG)
	puthex(fatInfo.fatStart);
	puts(" fatS");
#endif
	fatInfo.rootStart = GetWord(22);/*BPB_FATSz16*/
	if (fatInfo.rootStart == 0) {
	    if (!fatInfo.IS_FAT_32) 
		return 0x0b; /* should be FAT32; can not find root directory */
	    fatInfo.rootStart = GetLong(36);/*BPB_FATSz32*/
	}
	/* BPB_NumFATs * BPB_BlksPerSec fits in a word (8+7 bits) */
	fatInfo.rootStart = fatInfo.rootStart * (GetByte(16)/*BPB_NumFATs*/ *
						 BPB_BlksPerSec)
	    + fatInfo.fatStart;
#if 0 && defined(USE_DEBUG)
	puthex(fatInfo.rootStart);
	puts(" rootS");
#endif
	fatInfo.dataStart = fatInfo.rootStart + (fatInfo.BPB_RootEntCnt >> 4)
	    - (fatInfo.fatSectorsPerCluster << 1); /*first cluster is cluster 2*/

#if 0 && defined(USE_DEBUG)
	puthex(fatInfo.dataStart);
	puts(" dataS");
//	puthex(fatInfo.clusters);
//	puts(" Clustrs");
#endif
    }
    return 0;
}
#endif

#if 0
/** FAT/VFAT directory record union */
union DirRecordUnion {
  /** Standard directory entry */
  struct Entry {
    byte Name[11]; 0
    byte Attr;     11
    byte NTRes;    12
    byte CrtTimeTenth; 13
    word CrtTime;  14
    word CrtDate;  16
    word LstAccDate; 18
    word FstClusHi; 20
    word WrtTime;   22
    word WrtDate;   24
    word FstClusLo; 26
    u_32 FileSize;  28
  } entry;

  /** Extended directory entry */
  struct LongEntry {
    byte Ord;      0
    word Name1[5]; 1 /**< characters 1-5 */
    byte Attr;     11
    byte Type;     12 /**< entry type, zero=long name component */
    byte Chksum;   13
    word Name2[6]; 14 /**< characters 6-11 */
    word FstClusLO; 26 /**< zero for long entry */
    word Name3[2]; 28 /**< characters 12-13 */
  } longentry;
};
#endif


#define LAST_FRAGMENT 0x80000000UL
struct FRAGMENT {
    u_int32 start; /*high bit set if last fragment*/
    u_int16 size;
};
//#define MAX_FRAGMENTS 28//16//22
__y struct FRAGMENT fragments[MAX_FRAGMENTS];

#if 1
auto __y struct FRAGMENT *FragmentList(register __i2 __y struct FRAGMENT *frag,
			  register __b u_int32 fatCluster);
#else
auto __y struct FRAGMENT *FragmentList(register __i2 __y struct FRAGMENT *frag,
			  register __b u_int32 fatCluster) {
    register __d0 u_int16 offset;

    /* fatCluster means now current cluster, starting from 0, not 2. */
    /* note that startcluster=2 is already compensated in fatInitGlobals */
#if 0 && defined(USE_DEBUG)
    puthex(fatCluster>>16);
    puthex(fatCluster);
    puts(" fatC");
#endif
#if 0
    frag->size = 0;
    frag->start = fatCluster * fatInfo.fatSectorsPerCluster + fatInfo.dataStart;
#endif
    goto l55; /* read first sector */

    while (1) {
	u_int32 t;
	if (fatInfo.IS_FAT_32) {
	    t = GetLong(offset) & 0x0fffffffUL;
	} else {
#if 0 && defined(USE_DEBUG)
	    {
		register int i;
		for (i=0;i<256;i++) {
		    t = GetByte(i);
		    puthex(t);
		    putchar(' ');
		    if ((i&7)==7)
			putchar('\n');
		}
	    }
#endif

	    t = GetWord(offset);
	    if (t == 0xffffU)
		t = 0x0fffffffUL;
	}
#if 0 && defined(USE_DEBUG)
	puthex(t>>16);
	puthex(t);
	puts(" c");
#endif
	frag->size += fatInfo.fatSectorsPerCluster;
	++fatCluster;
	if (frag->size >= 32768U || t != fatCluster) {
	    /* if not sequential cluster, start next fragment */
#if 0 && defined(USE_DEBUG)
	    puts(" frag");
#endif
	    fatCluster = t;
	    if (frag == &fragments[MAX_FRAGMENTS-1] ||
		t == 0x0fffffffUL) {
		frag->start |= LAST_FRAGMENT;
		return frag+1;
	    }
	    frag++;
	l55:
	    frag->size = 0;
	    frag->start =
		fatCluster * fatInfo.fatSectorsPerCluster + fatInfo.dataStart;

	} else {
	    offset += (fatInfo.IS_FAT_32 ? 4 : 2);
	    offset &= 511;
	    if (offset != 0)
		continue;
	}
read:
	/* If FAT page is not already in memory, load it. */
	if (fatInfo.IS_FAT_32) {
	    offset = ((u_int16)fatCluster & 0x7F) << 2;
	    ReadDiskSector((fatCluster >> 7) + fatInfo.fatStart);
	} else {
	    offset = ((u_int16)fatCluster & 0xff) << 1;
	    ReadDiskSector((fatCluster >> 8) + fatInfo.fatStart);
	}
    }
}
#endif

u_int16 gFileNum[2];

#ifdef __VSDSP__
auto s_int16 OpenFile(register __c0 u_int16 fileNum);
#else
auto s_int16 OpenFile(register __c0 u_int16 fileNum) {
    register __i1 __y struct FRAGMENT *curFragment = &fragments[0];
    register __i0 __y struct FRAGMENT * __y nextFragment;

    /* Start at the start of root directory. */

    if (fatInfo.IS_FAT_32) {
	nextFragment = FragmentList(curFragment, 2);
    } else {
	curFragment->start = fatInfo.rootStart | LAST_FRAGMENT;
	curFragment->size  = (fatInfo.BPB_RootEntCnt >> 4); /* /sectorsPerCluster?*/
	nextFragment = curFragment + 1;
    }
    gFileNum[0] = fileNum;
    gFileNum[1] = 0;
    return HandleDir(curFragment, nextFragment);
}
#endif

#ifdef CHECKFILETYPE
#if 0
#define MKID(a,b,c) ((a)|((b)<<8)|((u_int32)(c)<<16))
auto s_int16 CheckFileType(register __a u_int32 name) {
    name &= 0x00ffffffUL;
    if (name == MKID('W','A','V')) {
	return -1;
    }
    if (name == MKID('M','P','3')) {
	return -1;
    }
#ifdef VS1003
    if (name == MKID('M','I','D')) {
	return -1;
    }
    if (name == MKID('W','M','A')) {
	return -1;
    }
    if (name == MKID('A','S','F')) {
	return -1;
    }
#endif
#ifdef VS1033
    if (name == MKID('M','I','D')) {
	return -1;
    }
    if (name == MKID('W','M','A')) {
	return -1;
    }
    if (name == MKID('A','S','F')) {
	return -1;
    }
    if (name == MKID('A','A','C')) {
	return -1;
    }
    if (name == MKID('M','P','4')) {
	return -1;
    }
    if (name == MKID('M','4','A')) {
	return -1;
    }
#endif
    return 0;
}
#endif/*1*/
#endif/*CHECKFILETYPE*/

/*147 words*/
auto s_int16 HandleDir(register __i2 __y struct FRAGMENT *curFragment,
		       __y struct FRAGMENT *nextFragment) {
    register __c1 int i;
    register u_int32 currentSector;

    currentSector = curFragment->start /*& 0x7fffffffUL */;
    i = 0;
#pragma msg 36 off
    goto first;
#pragma msg 36 on
    while (1) {
	register __d1 u_int16 fn = GetByte(i+0); /* first char of filename */

	/* We are now looking at FAT directory structure. */
	/* Is current file a regular file? */
	if (fn == 0)
	    goto eod; /*end of directory*/
	if (fn != 0xe5) { /* Not deleted */
#ifdef WITH_SUBDIRECTORIES
	    /* Is it a subdirectory? */
	    if ((GetByte(i+11/*Attr*/) & 0x10) && fn != '.') {
		/* TODO: do not call if FAT12! */
		if (HandleDir(nextFragment,
			      FragmentList(nextFragment,
					   ((u_int32)GetWord(i+20)<<16)+GetWord(i+26))) < 0)
		    return -1;
		ReadDiskSector(currentSector);/*reread sector*/
	    } else /* it was a subdirectory */
#endif
	    /* Attributes: NO directory, NO volume id, NO system, NO hidden */
	    if ((GetByte(i+11/*Attr*/) & 0xde) == 0
#ifdef CHECKFILETYPE
		/*
		  check supported files from name suffix:
		  vs1011/vs1002:  wav mp3
		  vs1003:         wav mp3 mid wma asf
		  vs1033: wav mp3 mid wma asf mp4 aac m4a
		*/
		&& CHECKFILETYPE(GetLong(i+8))
#endif
		) {

		/* It is a regular file. */

#if 0
		gFileNum[1]++;
#else
		{ /* force gFileNum[1]++ to a different register */
		    register u_int16 t = gFileNum[1] + 1;
		    gFileNum[1] = t;
		}
#endif
		if (gFileNum[0]-- == 0) {
#ifdef SCI_UI
		    /* Put the 8.3-character filename to memory so the user
		       can read it. Two bytes per word, big-endian. */
		    {   /* 13 words */
			register int j;
			register u_int16 *p = stream_buffer + i/2;
			register __y u_int16 *d = fileName;
			for (j=0;j<6;j++) {
			    *d++ = *p++;
			}
		    }
#endif
		    /* ------------ FILE FOUND ------------- */
		    fatInfo.fileSize = GetLong(i+28); /*do first!*/
		    FragmentList(&fragments[0],
				 ((u_int32)GetWord(i+20)<<16) + GetWord(i+26));
		    return -1; /* File found, All OK return */
		}
	    } /* normal file */
	} /* 0xe5 / 0x00 deleted or unused entry */
	i = (i + 32) & 511;
	if (i == 0) {
	    currentSector++;
	    if (--(curFragment->size) == 0) {
		if ((s_int16)(curFragment->start>>16) < 0) {
	eod:
		    return (s_int16)gFileNum[1];
		}
		curFragment++;
		currentSector = curFragment->start /*& 0x7fffffffUL*/;
	    }
	first:
#if 0 && defined(USE_DEBUG)
	    puthex(currentSector);
	    puts(" read");
#endif
	    ReadDiskSector(currentSector);
	}
    }
}


#ifdef HDATCHECK
u_int16 hdatCheck;
#endif
extern __y u_int16 hwSampleRate;

#if !defined(ENABLE_LAYER12) && (defined(VS1003) || defined(VS1033) || defined(VS1103) || defined(VS1053)) && !defined(NO_FAST_PAUSE)
#define FAST_PAUSE /* 'steal' already fetched stream data */
/* Can be done for all formats, because IdleHook is run as the main thread
   and not in interrupts. Whenever IdleHook is called, there either is
   not enough stream data anyway, or the decoder is outputting samples
   and there is not enough space in the audio buffer (thus not reading
   stream). */
#endif

#ifdef FAST_PAUSE
u_int16 __x * __x stream_save;
#endif

#ifndef SCI_UI
__y struct {
    u_int16 Data; /* must be first entry! Used by c.s */
    u_int16 Cnt;
    u_int16 Strobe;
    u_int16 Timer; /* Data,Cnt,Strobe,Timer must be in this order (buts.s) */
    u_int16 Key;
    s_int16 KeyTime;
} ui;
#endif

#if 1

#ifdef ENABLE_UI
s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
		 register __a0 s_int16 n);
#endif/*ENABLE_UI*/
#else/*1*/
#ifdef ENABLE_UI
s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
		 register __a0 s_int16 n) {
    /*Don't check mode, all we get is APPL_AUDIO.*/
    //if (mode == APPL_AUDIO)
    {
	register u_int16 t = hwSampleRate/16;
	/* generate a timebase for User Interface */
	ui.Cnt += n;
	if (ui.Cnt > t) {
	    ui.Cnt -= t;
	    ui.Timer++;
	    ui.Strobe = 1;
	}
    }
}
#endif/*ENABLE_UI*/
#endif

#if 1
auto u_int32 NextSector(void);
#else
auto u_int32 NextSector(void) {
    spix.sector++;
    if (spix.fragSize-1 == 0) {
	if ((s_int32)spix.sector < 0) /* was last fragment */
	    return -1;
	spix.fragment++;
	spix.sector  = spix.fragment->start /*& 0x7fffffffUL*/;
	spix.fragSize = spix.fragment->size; /* *fatInfo.sectorsPerCluster*/
    } else {
	spix.fragSize = spix.fragSize-1;
    }
    return 0;
}
#endif


u_int16 ClockSdi(void); /* dummy return value */
#if 0
u_int16 ClockSdi(void) {
    register int i;
    for (i=0;i<32;i++) {
	USEX(SER_DREQ) = 1;
	USEX(SER_DREQ) = 0;
    }
    return -1;
}
#endif

#if defined(FIX_WAV_DECODETIME) && defined(SCI_UI)
u_int16 lastDecodeTime;
#endif /*FIX_WAV_DECODETIME + SCI_UI*/

#ifdef USE_HC
s_int16 hcShift;
#endif

void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
#ifdef SAVE_VOLUME
int volChanged = 0;
#endif

#ifdef SAVE_STUFF
#define SPI_WREN  0x06
#define SPI_WRDI  0x04
#define SPI_RDSR  0x05
#define SPI_WRSR  0x01
#define SPI_READ  0x03
#define SPI_WRITE 0x02

register __a1 short SpiDelay(void);
auto u_int16 SpiSendReceiveSpi(register __a0 u_int16 datTopAligned, register __a1 s_int16 bits);

#if 1
void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
#else
void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value) {
    USEX(GPIO_ODATA) &= ~(SPI_xCS2|SPI_xCS);
    SpiSendReceiveSpi(SPI_WREN<<8, 8);
    USEX(GPIO_ODATA) |= SPI_xCS;
    USEX(GPIO_ODATA) &= ~SPI_xCS;
    SpiSendReceiveSpi(SPI_WRITE<<8, 8);
    SpiSendReceiveSpi(addr/*volume=26*/, 16);
    SpiSendReceiveSpi(value, 16);
    USEX(GPIO_ODATA) |= SPI_xCS|SPI_xCS2;//SPI_CLK;
}
#endif
//SaveWord(26/*volume=26*/, USEX(SCI_VOL));
#endif

#ifdef SHUFFLE_PLAY
u_int16 shuffle_offset = 0;
#endif

/*
  Called before HALT
 */
void UserHook(void) {
    register __c1 SPISTATE cachedSpiState = spiState;

#ifdef SCI_UI
    /* play pause/continue */
    if ((USEX(SCI_AICTRL3) & CTRL3_PAUSE_ON)) {
	if (cachedSpiState != spi_pause) {
	    oldSpiState = cachedSpiState;
	    cachedSpiState = spi_pause;
#if defined(FAST_PAUSE)
	    stream_save = stream_wr_pointer;
	    stream_wr_pointer = stream_rd_pointer+1;
	    /* Does not matter if it is one off..
	       StreamDiff() returns the right value regardless. */
#endif/*FAST_PAUSE*/
	}
    } else {
	if (cachedSpiState == spi_pause) {
	    cachedSpiState = oldSpiState;
#if defined(FAST_PAUSE)
	    stream_wr_pointer = stream_save;
#endif
	}
    }
    if ((USEX(SCI_AICTRL0) & 0x8000U)) {
	/* new song */
	USEX(SCI_AICTRL0) &= 0x7fff; /* ack command */
	/*NEW: clear pause and file ready bits automatically */
	USEX(SCI_AICTRL3) &= ~(CTRL3_PAUSE_ON | CTRL3_FILE_READY | CTRL3_AT_END);
	SpiSendClocks();

	/* Discard any unread data, including CRC, before starting again */
	{
	    register int i;
	    for (i=0; i<256+2; i++)
		SpiSendReceiveMmc(-1/*0xffff*/,16);
	}
	/* Jump to new track immediately */
	MyReset();
    }
#endif/*SCI_UI*/

#define SW1 0xffffU
#define SW2 0x8000U
#define SW3 0x7fffU

#ifdef ENABLE_UI
    if (ui.Strobe) {
	/* toggle MOSI to clock SDI */
	USEX(INT_ENABLE) = ENABLE_SCI_SDI_DAC_RX;
	ui.Data = ui.Strobe = 0;
	ClockSdi();
	if (ui.Data == 0) {
	    register u_int16 t = USEX(GPIO_ODATA), u;
	    USEX(GPIO_ODATA) = t & ~SPI_xCS;
	    ClockSdi();
	    u = ui.Data;
	    USEX(GPIO_ODATA) = t;
	    ui.Data = u>>1;/* 0xffff or 0x8000 ? -> 0x7fff 0x4000 */
	}

	USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;

#ifdef KEEP_SW2_PRESSED_PAUSES /*SW2 pauses until released*/
	if (ui.Data == SW2) {
	    if (cachedSpiState != spi_pause) {
		oldSpiState = cachedSpiState;
		cachedSpiState = spi_pause;
#if defined(FAST_PAUSE)
		stream_save = stream_wr_pointer;
		stream_wr_pointer = stream_rd_pointer+1;
		/* Does not matter if it is one off..
		   StreamDiff() returns the right value regardless. */
#endif/*FAST_PAUSE*/
	    }
	} else {
	    if (cachedSpiState == spi_pause) {
		cachedSpiState = oldSpiState;
#if defined(FAST_PAUSE)
		stream_wr_pointer = stream_save;
#endif
	    }
#endif/*KEEP_SW2_PRESSED_PAUSES*/

	if (ui.Data == ui.Key) {
	    if (ui.KeyTime < 8) {
		ui.KeyTime++;
	    } else if (ui.Key
#ifdef WAIT_FOR_BUTTON
		       && ui.KeyTime > 0
#endif
		) {
		/* long presses */
#ifdef NO_VOLUME_CONTROLS
		if (ui.Key == SW1 || ui.Key ==
#ifdef KEEP_SW2_PRESSED_PAUSES /*SW2 pauses until released*/
		    SW3
#else
		    SW2
#endif
		    ) {
		    /* Nothing done for long presses */
		} else
#else /*NO_VOLUME_CONTROLS*/
		register u_int16 vol = USEX(SCI_VOL);
		if (ui.Key == SW1) {
		    if (vol) {
			vol -= 0x0101;
			goto setvol;
		    }
		} else if (ui.Key ==
#ifdef KEEP_SW2_PRESSED_PAUSES /*SW2 pauses until released*/
			   SW3
#else
			   SW2
#endif
		    ) {
		    if (vol != (s_int16)0xffffU)
			vol += 0x0101;
		setvol:
		    USEX(SCI_VOL) = vol; /*in VS1011E disables ints! */
#ifdef VS1011E
		    USEX(INT_GLOB_ENA) = vol; /* enable interrupts, f*ck! */
		    /* also enables interrupts */
		    /*USEX(SCI_AICTRL0) = USEX(SCI_AICTRL0);*/
#endif
#if !defined(VS1053) && !defined(VS1033D)
		    SetVolumeFromSpi();
#endif
#ifdef SAVE_VOLUME
		    volChanged = 1;
#endif/*SAVE_VOLUME*/
		} else
#endif /*elseNO_VOLUME_CONTROLS*/
		{
#ifndef KEEP_SW2_PRESSED_PAUSES /*SW2 pauses until released*/
		    if (ui.KeyTime == 8) {
			ui.KeyTime = 9;
#if defined(LOUDNESS_ALWAYS_ON_OFF) && !defined(NO_RANDOM)
			/* if pause on -- toggle random play */
			USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
			if (cachedSpiState == spi_pause) {
			    cachedSpiState = oldSpiState;
#if defined(FAST_PAUSE)
			    stream_wr_pointer = stream_save;
#endif
			}

#else /*LOUDNESS_ALWAYS_ON_OFF*/

#if defined(VS1103)
			/* in VS1103 AICTRL2 is used for number of songs,
			   thus there is no loudness. */
			USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
#else/*VS1103*/
#ifdef NO_RANDOM
			/* toggle loudness - have LSb set when ON */
			USEX(SCI_BASS) ^= USEX(SCI_AICTRL2);
#if !defined(VS1053) && !defined(VS1033D)
			SetVolumeFromSpi();
#endif/*!VS1053*/
#else/*NO_RANDOM*/
			if (cachedSpiState
#ifdef SW3_RANDOM_PLAY
			    !=
#else
			    ==
#endif
			    spi_pause) {
			    /* if pause on -- toggle random play */
			    USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
			} else {
			    /* toggle loudness - have LSb set when ON */
			    USEX(SCI_BASS) ^= USEX(SCI_AICTRL2);
#if !defined(VS1053) && !defined(VS1033D)
			    SetVolumeFromSpi();
#endif
			}
#endif/*else NO_RANDOM*/
#endif/*else LOUDNESS_ALWAYS_ON_OFF*/
#endif/*elseVS1103*/
		    }
#endif /*!KEEP_SW2_PRESSED_PAUSES*/
		}
	    }
	} else {
	    if (ui.Data == 0
#ifdef WAIT_FOR_BUTTON
		&& ui.KeyTime > 0
#endif
		) {
		/* short presses */
		if (ui.KeyTime < 8) {
		    /* was pressed < 0.5 seconds */
#ifndef KEEP_SW2_PRESSED_PAUSES /*SW2 pauses until released*/
		    if (ui.Key == SW3) {
			/* play pause/continue -- not very good for midi.. */
			if (cachedSpiState != spi_pause) {
#ifndef UI_PAUSE_BEFORE_PLAY
			    oldSpiState = cachedSpiState;
			    cachedSpiState = spi_pause;
#if defined(FAST_PAUSE)
			    stream_save = stream_wr_pointer;
			    stream_wr_pointer = stream_rd_pointer+1;
#endif/*FAST_PAUSE*/
#endif/*!UI_PAUSE_BEFORE_PLAY*/
			} else {
			    cachedSpiState = oldSpiState;
#if defined(FAST_PAUSE)
			    /* restore 'stolen' data */
			    stream_wr_pointer = stream_save;
#endif
			}
		    } else
#endif /*!KEEP_SW2_PRESSED_PAUSES*/
		    if (ui.Key == SW1 || ui.Key ==
#ifdef KEEP_SW2_PRESSED_PAUSES /*SW2 pauses until released*/
			SW3
#else
			SW2
#endif
			) {
			register u_int16 song = USEX(SCI_AICTRL0);
			if (ui.Key == SW1) {
			    song++; /* next file */
			} else
#ifndef UI_PAUSE_BEFORE_PLAY
			    if (USEX(SCI_DECODE_TIME) < 5)
#endif
			    {
			    if (song == 0) {
#ifdef VS1103
				song = USEX(SCI_AICTRL2)-1;
#else
				song = USEX(SCI_AICTRL1)-1;
#endif
			    } else { /* previous file */
				song--;
			    }
			    /* else the same file */
			}
#ifdef VS1011E
			/* write to SCI_AICTRL0 enables interrupts,
			   but they are enabled anyway.. */
#endif
			USEX(SCI_AICTRL0) = song;

			/* especially needed for MIDI!! */
#if !defined(NO_MIDI) && (!defined(ENABLE_LAYER12) || !defined(HDATCHECK))
			USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
#endif
			cachedSpiState = spi_zeros;
#if defined(ENABLE_LAYER12)
			/* gain=0 for partial last frame (bugs) */
			volume[0] = volume[1] = 0;
#endif
		    }
		}
#if defined(VS1011E) && defined(SAVE_VOLUME)
		/*Save volume when volume button is released.
		  Only works for cards supporting interrupted CS.
		  Playback may fail with cards not supporting
		  interrupted CS.
		  With VS1011E this is not critical because
		  there is only mp3 and wav.
		*/
		else if (volChanged) {
		    volChanged = 0;
		    SaveWord(26/*volume=26*/, USEX(SCI_VOL));
		}
#endif/*SAVE_VOLUME*/

	    }
	    ui.Key = ui.Data;
	    ui.KeyTime = 0;
	}
#ifdef KEEP_SW2_PRESSED_PAUSES /*SW2 pauses until released*/
	}
#endif/*KEEP_SW2_PRESSED_PAUSES*/

#if !defined(VS1011E) && defined(SAVE_VOLUME)
	/*Save volume when volume button is released. */
	if (ui.Data == 0 &&
	    volChanged &&
	    cachedSpiState == spi_seek) {
	    volChanged = 0;
	    SaveWord(26/*volume=26*/, USEX(SCI_VOL));
	}
#endif/*SAVE_VOLUME*/

	//ui.Data = 0;
    }
#endif /*USE_UI*/


#ifndef USE_DEBUG
    if (cachedSpiState == spi_file) {
	/* extra byte read if odd file size */
	spix.size = (fatInfo.fileSize+1)/2; /* in words! */
	cachedSpiState = spi_seek;
setfrag:
	spix.fragment = &fragments[0];
	spix.sector   = fragments[0].start /*& 0x7fffffffUL*/;
	spix.fragSize = fragments[0].size; /* *fatInfo.sectorsPerCluster*/
    }
#if defined(FIX_WAV_DECODETIME) && defined(SCI_UI)
    if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) == CTRL3_LOOP_SONG &&
	USEX(SCI_HDAT1) == 0x7665/*(('v'<<8)|'e')*/) {
	if (USEX(SCI_DECODE_TIME) < lastDecodeTime)
	    USEX(SCI_DECODE_TIME) = lastDecodeTime;
    }
#endif /*SCI_UI and FIX_WAV_DECODETIME*/

    if (cachedSpiState == spi_seek) {
#ifdef USE_HC
	MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,
		   (spix.sector & 0x7fffffffUL) << hcShift);
#else
	MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(spix.sector<<9));
#endif
	cachedSpiState = spi_waitdata;
	spiCnt = 0;
	NextSector();
#ifdef HDATCHECK
	if (++hdatCheck == HDATCHECK && USEX(SCI_HDAT1) == 0) {
	    /* if file not detected at HDATCHECK/2 kB, skip the rest */
	    goto end_of_file;
	}
#endif
    }
    if (cachedSpiState == spi_waitdata) {
	if (SpiSendReceiveMmc(0xff00,8) == 0xfe) {
	    cachedSpiState = spi_data;
	    spiCnt = 0;
#ifdef TIMEOUT
	    timeOut = 0;
#endif
	} else {
	    if ((s_int16)++spiCnt < 0)
		MyReset(); /* timeout */
	}
    } else if (cachedSpiState == spi_data) {
	while (StreamDiff() < STREAM_BUFFER_SZ-40 &&
	       spix.size > 0) {
	    register int i, j = 32;
	    register u_int16 *wr = (void *)stream_wr_pointer;
	    if (j > spix.size)
		j = spix.size;
	    for (i=0;i<j;i++) {
		*wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
		if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
		    wr = stream_buffer;
	    }
	    stream_wr_pointer = (void *)wr;
	    spix.size -= j;//32;
	    spiCnt += 32;
	    if (spiCnt >= 256) {
		/* 256 words read, get next block in file */
		cachedSpiState = spi_seek;
		SpiSendReceiveMmc(-1/*0xffff*/,16); /* discard crc */
		break;
	    }
#ifdef TIMEOUT
	    timeOut = 0;
#endif
	    /* watchdog clearing here doesn't work with midi */
	}

#ifdef TIMEOUT
	if ((s_int16)++timeOut == -1) {
	    spix.size = 0;
	}
#endif

	if (spix.size <= 0) {
end_of_file:
	    SpiSendClocks();

#ifdef SCI_UI
	    if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) == CTRL3_LOOP_SONG){
		/* restart the same file immediately */
#if defined(FIX_WAV_DECODETIME) && defined(SCI_UI)
		lastDecodeTime = USEX(SCI_DECODE_TIME)+1;
#endif /*SCI_UI and FIX_WAV_DECODETIME*/
		cachedSpiState = spi_file;
	    } else
#endif/*SCI_UI*/
	    {
		USEX(SCI_AICTRL0)++; /* next file */
		cachedSpiState = spi_zeros;
#ifdef SCI_UI
		/*NEW: pause after play */
		if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
		    CTRL3_PAUSE_AFTER_PLAY) {
		    USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON | CTRL3_AT_END;
		}
#endif
	    }
	}
    } else if (cachedSpiState == spi_zeros) {
#if 0
	/* Discard any unread data, including CRC, before starting again */
	SpiSendReceiveMmc(-1/*0xffff*/,16);
#endif
	if (StreamDiff() < STREAM_BUFFER_SZ-40) {
	    register int i, z = 0;
	    register u_int16 *wr = (void *)stream_wr_pointer;
	    for (i=0;i<32;i++) {
		*wr++ = z;
		if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
		    wr = stream_buffer;
	    }
	    ++spiCnt;
	    stream_wr_pointer = (void *)wr;
	    /* watchdog clearing here doesn't work with midi */
#ifdef TIMEOUT
	    timeOut = 0;
#endif
	}
#ifdef TIMEOUT
	if ((s_int16)timeOut++ < 0) {
//putch(USEX(SCI_HDAT1)>>8);
//putch(USEX(SCI_HDAT1));
	    goto goreset;
	}
#endif
	if (spiCnt >= 256+2048/32) { /* was left to max 256 by spi_data */
	    /* 2048 words instead of bytes */
goreset:
	    /* Discard any unread data, including CRC, before starting again */
#if 1 /* 8 words */
	    {
		register int i;
		for (i=0; i<256+2; i++)
		    SpiSendReceiveMmc(-1/*0xffff*/,16);
	    }
#endif
	    SpiSendClocks();

	    MyReset();
	}
    } else {
    }

#ifdef SCI_UI
    USEX(SER_DREQ) = 0;
#else

#ifndef NO_DREQ_LED
    /* 15 words */
    {
	register int t = 0;

	if ((ui.Timer & 0x18) == 0) {
	    /* DREQ indicates random play status */
#ifdef NO_RANDOM
	    t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
#else
	    t = USEX(SCI_AICTRL3);  /* lowest bit is global random play */
#ifdef LOUDNESS_ALWAYS_ON_OFF
	    /* indicate random play */
#else /*LOUDNESS_ALWAYS_ON_OFF*/

	    if (cachedSpiState
#ifdef SW3_RANDOM_PLAY
		==
#else
		!=
#endif
		spi_pause)
		t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
#endif /*else LOUDNESS_ALWAYS_ON_OFF*/
#endif /*else NO_RANDOM*/
	}
	USEX(SER_DREQ) = t;
    }
#endif/*!NO_DREQ_LED*/
#endif/*!SCI_UI*/

#endif/*!USE_DEBUG*/
    spiState = cachedSpiState;

}


#if defined(ENABLE_LAYER12) && defined(MP2_PATCH)
void Mp2Patch(void);
#endif


auto void MyMain(register s_int16 __a1 fullReset) {
#if defined(SCI_UI)
    USEX(SCI_AICTRL1) = 0;
#endif
    USEX(SCI_MODE) = SciModeOr() | (USEX(SCI_MODE) & ~(1<<SCIMB_OUT_OF_WAV));

    USEX(GPIO_ODATA) = SPI_xCS|SPI_xCS2|SPI_CLK; /*use same= -1 word*/
    USEX(GPIO_DDR)   = SPI_xCS|SPI_CLK|SPI_xCS2;


    if (fullReset)
#ifdef NO_RANDOM
	InitHardware(); /* clears memory, so can't be used for debugging */
#else
    {
	register int r = rand_seed;
	InitHardware(); /* clears memory, so can't be used for debugging */
	rand_seed = r;
    }
#endif
    else {
	//hwSampleRate = 8888;
	SetClockFromSpi();
	//stream_diff = 0;
	stream_rd_odd = 0;
	stream_wr_pointer = stream_rd_pointer = (__x u_int16 *)stream_buffer;
	audio_wr_pointer = audio_rd_pointer = &audio_buffer[0];
	ymemset((__y void *)audio_buffer, 0, AUDIO_BUFFER_SZ);
    }

    InitLayer3();
    InitCommon();
    InitDecode();

    if (fullReset) {
	SetHardware(1, 500);
	/*otherwise keep it in powerdown until MMC found*/
	USEX(SCI_STATUS) |= (1<<SCIST_APDOWN1)|(1<<SCIST_APDOWN2);
    } else {
	/* If not powerup reset, re-enable analog immediately. */
	USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN2);
    }

#ifdef ENABLE_UI
    applAddr = ApplAddr;
#endif/*ENABLE_UI*/

    /* Allow DREQ and enable interrupts, VERY IMPORTANT! */
//    USEX(INT_ENABLE) = 0;
//    USEX(INT_GLOB_ENA) = 0;
    USEX(INT_GLOB_ENA) = 0;
//    USEX(INT_GLOB_ENA) = 0;

    InitReadFrame();

    /* wake up MMC! -- If this takes too long, watchdog will bite. */
    //while (1)
    {
	register s_int16 i, cmd;
    tryagain:
	for (i=0; i<512; i++) {
	    SpiSendClocks();
	}
	/* MMC Init, command 0x40 should return 0x01 if all is ok. */
	i = MmcCommand(MMC_GO_IDLE_STATE/*CMD0*/|0x40,0);
	if (i != 1)
	    goto tryagain;//continue; /* No valid idle response */
	cmd = MMC_SEND_OP_COND|0x40;
#ifdef USE_HC
	//MmcCommand(MMC_CRC_ON_OFF/*CMD59*/|0x40, 0);
	if (MmcCommand(MMC_SEND_IF_COND/*CMD8*/|0x40, 0x00000122/*2.7-3.6V*/)
	    == 1) {
	    /* MMC answers: 0x05 SD answers: 0x01 */
	    /* Should we read the whole R7 response? */
	    //SpiSendReceiveMmc(-1, 32);
	    cmd = 0x40|41; /* ACMD41 - SD_SEND_OP_COND */
	}
#endif /*USE_HC*/
	/* MMC Wake-up call, set to Not Idle (mmc returns 0x00)*/
	//i = 0; /* i is 1 when entered .. but does not make the code shorter*/
	while (1) {
	    register int c;
#ifdef USE_HC
	    if (cmd == (0x40|41)) {
		MmcCommand(0x40|55/*CMD55*/,0);
		c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/, 0x40000000UL);
	    } else
#endif
	    c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/, 0);
	    if (c == 0)
		break;
	    if (++i >= 12000/*< 0*/ || c != 1) {
#ifdef USE_DEBUG
		puts("again");
#endif
		goto tryagain; /* Not able to power up mmc */
	    }
	}

#ifdef USE_HC
	hcShift = 9;
	i = 0x0246;
	/* PRETEC's 1G MMC does not seemd to handle CMD58 here! */
	/* Support HC for SD, but not for MMC! */
	if (cmd == (0x40|41) &&
	    MmcCommand(MMC_READ_OCR/*CMD58*/|0x40, 0) == 0) {
	    if (SpiSendReceiveMmc(-1, 16) & (1<<(30-16))) {
		/* OCR[30]:CCS - card capacity select */
		/* HighCapacity! */
		/* ldc 9,D0 (0x0246) <-> ldc 0,D0 (0x0006) */
		i = 6;
		hcShift = 0;
	    }
	    //SpiSendReceiveMmc(-1, 16);
	}
	WriteToProgramRam((u_int16)ReadDiskSector+2, 0, i);
#endif /*USE_HC*/

#if 0
	/* Set Block Size of 512 bytes -- default for at least HC */
	/* Needed by MaxNova S043618ATA 2J310700 MV016Q-MMC */
	/* Must check return value! (MicroSD restart!) */
	if (MmcCommand(MMC_SET_BLOCKLEN|0x40, 512) != 0)
	    goto tryagain;
	/* All OK return */
#endif
    }


    if (InitFileSystem()  /* Detects cards without partition table */
	== 0) {

#ifdef SCI_UI
	if ((USEX(SCI_AICTRL3) & CTRL3_NO_NUMFILES)) {
	    USEX(SCI_AICTRL1) = 0x7fff;
	} else
#endif/*SCI_UI*/
	{
	    /* determine number of playable files */
	    USEX(SCI_AICTRL1) = OpenFile(0xffffU);
	}

#if !defined(NO_RANDOM) /* no random play feature */
	if ((USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY)) {
	    /* random play (not shuffle!) */
#ifdef SHUFFLE_PLAY /*shuffle +12/19 words*/
	    {
		register __b1 u_int16 files = USEX(SCI_AICTRL1);
		if (files >= 2) {
		    register __b0 u_int16 t;
		    do {
			t = myrand() + shuffle_offset;
			if (rand_seed == 0)
			    shuffle_offset += files;
//USEX(SER_DREQ) ^= 1;
		    } while (t >= files
#if 0
			     || t == USEX(SCI_AICTRL0)-1 /*not same twice*/
#endif
			);
		    USEX(SCI_AICTRL0) = t;
		}
	    }
#else
	    USEX(SCI_AICTRL0) = (myrand() % USEX(SCI_AICTRL1));
#endif
	}
#endif
	if (OpenFile(USEX(SCI_AICTRL0)
#ifdef SCI_UI
		      &= 0x7fffU
		     //max 32768 files in SCI-controlled version
#endif
	    ) == -1) {
#if defined(SKIP_ID3V2) /* 48 words */
	    /* check for ID3v2 */
	    /* An ID3v2 tag can be detected with the following pattern:
	       $49 44 33 yy yy xx zz zz zz zz */
	    ReadDiskSector(fragments[0].start);
	    if ((GetLong(0) & 0xffffffL) == 0x334449) {
		/* skip in 16kB resolution, assume tags are smaller than 2MB */
		/* 0.5kB resolution now */
		register u_int16 skip = 32 * GetByte(7) + GetByte(8)/4;
		/* guard against small fragments */
		if (skip >= fragments[0].size)
		    skip = fragments[0].size-1;
		fragments[0].start += skip;
		fragments[0].size  -= skip;
		fatInfo.fileSize -= (u_int32)skip * 512;
	    }
#endif

#if defined(SCI_UI)
	    if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
		CTRL3_PAUSE_BEFORE_PLAY) {
		/* Mark file ready and go to pause state in UserHook */
		USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
	    }
	    /*NEW: set FILE_READY always */
	    USEX(SCI_AICTRL3) = CTRL3_FILE_READY | 
		(USEX(SCI_AICTRL3) & ~CTRL3_AT_END);
#endif


#if defined(UI_PAUSE_BEFORE_PLAY) && defined(ENABLE_UI)
	    oldSpiState = spi_file;
	    spiState = spi_pause;
#if defined(FAST_PAUSE)
	    stream_save = stream_wr_pointer;
#endif
#else
	    spiState = spi_file;
#endif
#ifdef HDATCHECK
	    hdatCheck = 0;
#endif
	} else {
	gofirst:
	    USEX(SCI_AICTRL0) = 0; /* file not found, jump to first file */
	    MyReset();
	}
    } else {
	/* Initialize again */
	goto gofirst;//MyReset();
    }

#if !defined(SCI_UI) && defined(WAIT_FOR_BUTTON)
    if ((USEX(SCI_AICTRL3)&CTRL3_PLAY_MODE_MASK) == CTRL3_PAUSE_BEFORE_PLAY) {
	/* Only once per boot */
	USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
	while (1) {
	    ui.Data = 0;
	    /* toggle MOSI to clock SDI */
	    USEX(INT_ENABLE) = ENABLE_SCI_SDI_DAC_RX;
	    ClockSdi();
	    if (ui.Data == 0) {
		USEX(GPIO_ODATA) &= ~SPI_xCS;
		ClockSdi();
		USEX(GPIO_ODATA) |= SPI_xCS;
		ui.Data >>= 1;/* 0xffff or 0x8000 ? -> 0x7fff 0x4000 */
	    }
	    USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
	    if (ui.Data) {
		ui.Key = ui.Data;
		ui.KeyTime = -32767;
		break;
	    }
	    //Halt();//Sleep();
	}
    }
#endif


#if defined(KEEP_DRIVERS_OFF)
    /*otherwise keep it in powerdown until MMC found*/
    USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1); /* enable ANALOG power */
#endif


#ifdef SAVE_POSITION
    SaveWord(28/*AICTRL0*/, USEX(SCI_AICTRL0));
#endif


    USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
    USEX(SCI_DECODE_TIME) = 0;

#if defined(FIX_WAV_DECODETIME) && defined(SCI_UI)
    lastDecodeTime = 0;
#endif /*SCI_UI and FIX_WAV_DECODETIME*/

#if defined(ENABLE_LAYER12) && defined(MP2_PATCH)
    Mp2Patch();
#else
    while (1) {
	if (!ReadFrame()) {
	    play_frame();
	} else {
	    InitReadFrame();
	}
    }
#endif
}
